#!/usr/bin/env python
# -*- coding: utf-8 -*-


import json
import time
import unittest

import iland

import requests_mock


BASE_URL = 'http://example.com/ecs'

VALID_TOKEN_PAYLOAD = {'expires_in': 12,
                       'refresh_expires_in': 17,
                       'access_token': 'AZERTYUIOP',
                       'refresh_token': 'BLABLABLA'}

VALID_REFRESH_TOKEN_PAYLOAD = {'expires_in': 12,
                               'refresh_expires_in': 17,
                               'access_token': 'QWERTYUIOP',
                               'refresh_token': 'BLOBLOBLO'}


class TestIland(unittest.TestCase):
    session = None
    adapter = None

    def setUp(self):
        self.api = iland.Api(client_id='fake',
                             client_secret='fake',
                             username='fake',
                             password='fake')
        self.api._base_url = BASE_URL
        self.api._access_token_url = iland.ACCESS_URL

    def test_login_ok_200(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            self.api.login()
            self.assertEqual(VALID_TOKEN_PAYLOAD, self.api.get_access_token())

    def test_login_ok_201(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=201)
            self.api.login()
            self.assertEqual(VALID_TOKEN_PAYLOAD, self.api.get_access_token())

    def test_login_ok_202(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=202)
            self.api.login()
            self.assertEqual(VALID_TOKEN_PAYLOAD, self.api.get_access_token())

    def test_login_ko_500(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps({'error': 'an error occured'}),
                   status_code=500)
            with self.assertRaises(iland.UnauthorizedException):
                self.api.login()

    def test_login_ko_400(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps({'error': 'an error occured'}),
                   status_code=400)
            with self.assertRaises(iland.UnauthorizedException):
                self.api.login()

    def test_refresh_token_ok(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            m.post(iland.REFRESH_URL,
                   text=json.dumps(VALID_REFRESH_TOKEN_PAYLOAD),
                   status_code=200)
            self.api.login()
            self.assertEqual(VALID_TOKEN_PAYLOAD, self.api.get_access_token())

            # manually refresh token
            self.api.refresh_access_token()
            # still the same since not expired therefore not renewed
            self.assertEqual(VALID_TOKEN_PAYLOAD, self.api.get_access_token())

            # let's wait for expiration
            time.sleep(5)
            self.api.refresh_access_token()
            self.assertEqual(VALID_REFRESH_TOKEN_PAYLOAD,
                             self.api.get_access_token())

            # manually remove the actual token so that we refetch an access
            # token
            self.api._token = None
            self.api.refresh_access_token()
            self.assertEqual(VALID_TOKEN_PAYLOAD,
                             self.api.get_access_token())

    def test_refresh_token_ko_400(self):
        with requests_mock.mock() as m:
            login_spy = m.post(iland.ACCESS_URL,
                               text=json.dumps(VALID_TOKEN_PAYLOAD),
                               status_code=200)
            refresh_spy = m.post(iland.REFRESH_URL,
                                 text=json.dumps(VALID_REFRESH_TOKEN_PAYLOAD),
                                 status_code=400)
            self.api.login()
            self.assertEqual(VALID_TOKEN_PAYLOAD, self.api.get_access_token())
            # wait for access token expiration. since the refresh endpoint
            # returns a 400, we expect that a new login will be initiated
            # because the existing session can no longer be refreshed
            time.sleep(5)
            self.assertEqual(VALID_REFRESH_TOKEN_PAYLOAD,
                             self.api.refresh_access_token())
            self.assertEqual(1, refresh_spy.call_count)
            self.assertEqual(2, login_spy.call_count)

    def test_refresh_token_ko_500(self):
        with requests_mock.mock() as m:
            login_spy = m.post(iland.ACCESS_URL,
                               text=json.dumps(VALID_TOKEN_PAYLOAD),
                               status_code=200)
            refresh_spy = m.post(iland.REFRESH_URL,
                                 text=json.dumps(VALID_REFRESH_TOKEN_PAYLOAD),
                                 status_code=500)
            self.api.login()
            self.assertEqual(VALID_TOKEN_PAYLOAD, self.api.get_access_token())
            # wait for access token expiration. since the refresh endpoint
            # returns a 400, we expect that a new login will be initiated
            # because the existing session can no longer be refreshed
            time.sleep(5)
            self.assertEqual(VALID_REFRESH_TOKEN_PAYLOAD,
                             self.api.refresh_access_token())
            self.assertEqual(1, refresh_spy.call_count)
            self.assertEqual(2, login_spy.call_count)

    def test_refresh_token_expired(self):
        with requests_mock.mock() as m:
            login_spy = m.post(iland.ACCESS_URL,
                               text=json.dumps(VALID_TOKEN_PAYLOAD),
                               status_code=200)
            self.api.login()
            self.assertEqual(VALID_TOKEN_PAYLOAD, self.api.get_access_token())
            # wait for refresh token expiration. since the refresh token
            # expired, we expect that a new login will be initiated because
            # the existing session can no longer be refreshed
            time.sleep(8)
            self.assertEqual(VALID_TOKEN_PAYLOAD,
                             self.api.refresh_access_token())
            self.assertEqual(2, login_spy.call_count)

    def test_get_ok_200(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=200)
            req = self.api.get(rpath)
            self.assertEquals(user_data, req)

    def test_get_ok_201(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=201)
            req = self.api.get(rpath)
            self.assertEquals(user_data, req)

    def test_get_ok_202(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=202)
            req = self.api.get(rpath)
            self.assertEquals(user_data, req)

    def test_get_ok_204(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=204)
            req = self.api.get(rpath)
            self.assertEquals(user_data, req)

    def test_get_ko_400(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=400)
            with self.assertRaises(iland.ApiException):
                self.api.get(rpath)

    def test_get_ko_500(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=500)
            with self.assertRaises(iland.ApiException):
                self.api.get(rpath)

    def test_post_ok(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.post(BASE_URL + rpath, text=json.dumps(user_data),
                   status_code=200)
            req = self.api.post(rpath, form_data={'a': 'b'})
            self.assertEquals(user_data, req)

    def test_post_ok_no_formdata(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.post(BASE_URL + rpath, text=json.dumps(user_data),
                   status_code=200)
            req = self.api.post(rpath)
            self.assertEquals(user_data, req)

    def test_put_ok(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.put(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=200)
            req = self.api.put(rpath, form_data={'a': 'b'})
            self.assertEquals(user_data, req)

    def test_put_ok_no_formdata(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.put(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=200)
            req = self.api.put(rpath)
            self.assertEquals(user_data, req)

    def test_delete_ok(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.delete(BASE_URL + rpath, text=json.dumps(user_data),
                     status_code=200)
            req = self.api.delete(rpath)
            self.assertEquals(user_data, req)

    def test_unknown_verb_internal(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.delete(BASE_URL + rpath, text=json.dumps(user_data),
                     status_code=200)
            with self.assertRaises(iland.ApiException):
                self.api._do_request(rpath, verb='ACK')

    def test_with_default_base_url(self):
        self.api = iland.Api(client_id='fake',
                             client_secret='fake',
                             username='fake',
                             password='fake')
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(iland.BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=200)
            req = self.api.get(rpath)
            self.assertEquals(user_data, req)

    def test_with_proxies_set(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=200)
            self.api._proxies = {'https': 'https://10.10.10.10:3128'}
            req = self.api.get(rpath)
            self.assertEquals(user_data, req)

    def test_get_with_extra_header(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  request_headers={'Host': 'api.ilandcloud.com'},
                  status_code=200)
            req = self.api.get(rpath, headers={'Host': 'api.ilandcloud.com'})
            self.assertEquals(user_data, req)

    def test_get_with_extra_disallowed_header(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=200)
            # Set Accept to text/csv but it's ignored by api, so we get json
            req = self.api.get(rpath, headers={'Accept': 'text/csv'})
            self.assertEquals(user_data, req)

    def test_get_with_timeout(self):
        with requests_mock.mock() as m:
            m.post(iland.ACCESS_URL,
                   text=json.dumps(VALID_TOKEN_PAYLOAD),
                   status_code=200)
            rpath = '/user/jchirac'
            user_data = {'username': 'jchirac'}
            m.get(BASE_URL + rpath, text=json.dumps(user_data),
                  status_code=200)
            req = self.api.get(rpath, timeout=5.0)
            self.assertEquals(user_data, req)



# -*- coding: utf-8 -*-

"""A library that provides constants for the `iland.Api` object."""

#: Base URL to the iland cloud API.
BASE_URL = 'https://api.ilandcloud.com/v1'

#: Access token URL.
ACCESS_URL = \
    'https://console.ilandcloud.com/auth/realms/iland-core/' \
    + 'protocol/openid-connect/token'

#: Refresh token URL. (`refresh` query param is here only for mock
# testing reason)
REFRESH_URL = ACCESS_URL + '?refresh=1'



# -*- coding: utf-8 -*-

"""A module that provides a Python interface to the iland cloud API."""

from .api import Api  # noqa
from .constant import ACCESS_URL  # noqa
from .constant import BASE_URL  # noqa
from .constant import REFRESH_URL  # noqa
from .exception import ApiException  # noqa
from .exception import UnauthorizedException  # noqa

__author__ = 'iland Internet Solutions, Corp'
__email__ = 'devops@iland.com'
__version__ = '1.0.4'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_iland_int
----------------------------------

Integration tests for the `iland` module.
"""

import time
import unittest

import iland
import iland.constant
from iland.exception import ApiException, UnauthorizedException

try:
    # only PyCharm w/ Python3 fails here without dot
    from apicreds import (CLIENT_ID,
                          CLIENT_SECRET,
                          USERNAME,
                          PASSWORD,
                          PROXIES)
except ImportError:
    try:
        from .apicreds import (CLIENT_ID,
                               CLIENT_SECRET,
                               USERNAME,
                               PASSWORD,
                               PROXIES)
    except ImportError:
        CLIENT_ID = None
        CLIENT_SECRET = None
        USERNAME = None
        PASSWORD = None
        PROXIES = {}

VDC_UUID = \
    'res01.ilandcloud.com:urn:vcloud:vdc:a066325d-6be0-4733-8d9f-7687c36f4536'


@unittest.skipIf(not CLIENT_ID and not CLIENT_SECRET and not USERNAME and not
                 PASSWORD, "No credentials provided")
class TestIlandInt(unittest.TestCase):
    def setUp(self):
        self._api = iland.Api(client_id=CLIENT_ID,
                              client_secret=CLIENT_SECRET,
                              username=USERNAME,
                              password=PASSWORD)
        self._api._base_url = iland.constant.BASE_URL

    def tearDown(self):
        pass

    def test_get(self):
        user = self._api.get('/users/' + USERNAME)
        self.assertEqual(USERNAME, user.get('name'))
        self.assertTrue(len(user.keys()) > 5)

    def test_post(self):
        user_alert_emails = self._api.get('/users/' + USERNAME +
                                          '/alert-emails')
        self.assertTrue(len(user_alert_emails['emails']) >= 1)
        self.assertEquals(user_alert_emails['username'], USERNAME)

        old_user_alert_emails = user_alert_emails
        user_alert_emails = {'emails': ['test@iland.com', 'test2@iland.com'], }
        self._api.post('/users/' + USERNAME + '/actions/update-alert-emails',
                       user_alert_emails)

        self.assertEqual(2, len(user_alert_emails['emails']))

        self._api.post('/users/' + USERNAME + '/actions/update-alert-emails',
                       old_user_alert_emails)
        self.assertEquals(len(user_alert_emails['emails']), 2)

    def test_put_delete(self):
        vdc_uuid = VDC_UUID
        vdc_md = self._api.get('/vdcs/' + vdc_uuid + '/metadata')
        self.assertEquals(len(vdc_md['data']), 0)

        new_md = [{'type': 'string',
                   'value': 'B',
                   'access': 'READ_WRITE',
                   'key': 'AAA'}, ]
        new_md.extend(vdc_md['data'])
        self._api.put('/vdcs/' + vdc_uuid + '/metadata', form_data=new_md)

        time.sleep(10)

        updated_md = self._api.get('/vdcs/' + vdc_uuid + '/metadata')
        self.assertEquals(len(updated_md['data']), 1)

        self._api.delete('/vdcs/' + vdc_uuid + '/metadata/' + 'AAA')

        time.sleep(10)

        updated_md = self._api.get('/vdcs/' + vdc_uuid + '/metadata')
        self.assertEquals(len(updated_md['data']), 0)

    def test_get_with_host_header(self):
        user = self._api.get('/users/' + USERNAME,
                             headers={'Host': 'api.ilandcloud.com'})
        self.assertEqual(USERNAME, user.get('name'))
        self.assertTrue(len(user.keys()) > 5)

    def test_get_with_disallowed_header(self):
        user = self._api.get('/users/' + USERNAME,
                             headers={'Accept': 'text/csv'})
        self.assertEqual(USERNAME, user.get('name'))
        self.assertTrue(len(user.keys()) > 5)

    def test_refresh_token(self):
        self._api.login()
        self.assertIsNotNone(self._api.get_access_token())
        expires_in = self._api._token_expiration_time

        # force expires for tests.
        self._api._token_expiration_time = 0
        self._api.refresh_access_token()

        new_expires_in = self._api._token_expiration_time
        self.assertTrue(new_expires_in > expires_in)

    def test_unauthorized_errors(self):
        wrongCredsApi = iland.Api(client_id=CLIENT_ID,
                                  client_secret=CLIENT_SECRET,
                                  username='PYTHON_SDK_TEST',
                                  password='XXXX')
        wrongCredsApi._base_url = iland.constant.BASE_URL

        with self.assertRaises(UnauthorizedException):
            wrongCredsApi.login()

        with self.assertRaises(UnauthorizedException):
            wrongCredsApi.get_access_token()

        with self.assertRaises(UnauthorizedException):
            wrongCredsApi.refresh_access_token()

    def test_api_errors(self):
        with self.assertRaises(ApiException):
            self._api.get('/doesnotexist')

    @unittest.skipIf(not PROXIES, "No proxies defined")
    def test_get_with_proxy(self):
        self._api._proxies = PROXIES
        user = self._api.get('/users/' + USERNAME)
        self.assertEqual(USERNAME, user.get('name'))
        self.assertTrue(len(user.keys()) > 5)



# -*- coding: utf-8 -*-

"""A library that provides custom exceptions for the `iland.Api` module."""


class ApiException(Exception):
    """Base class for `iland.Api` errors."""

    @property
    def error(self):
        """Returns the first argument used to construct this error."""
        return self.args[0]

    @property
    def message(self):
        """Returns the second argument used to construct this error."""
        return self.args[1]

    @property
    def detail_message(self):
        """Returns the third argument used to construct this error."""
        return self.args[2]


class UnauthorizedException(ApiException):
    """ Unauthorized `iland.Api` exception. """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from setuptools import find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.read().splitlines()

with open('requirements_dev.txt') as requirements_dev_file:
    test_requirements = requirements_dev_file.read().splitlines()

setup(
    name='iland-sdk',
    version='1.0.4',
    description="iland cloud Python SDK.",
    long_description=readme + '\n\n' + history,
    author="iland Internet Solutions, Corp",
    author_email='devops@iland.com',
    url='https://github.com/ilanddev/python-sdk',
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='iland cloud sdk',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-

"""A library that provides a Python interface to the iland cloud API."""

import json
import time

import requests

from .constant import ACCESS_URL, BASE_URL, REFRESH_URL
from .exception import ApiException, UnauthorizedException
from .log import LOG


class Api(object):
    """A Python interface into the iland cloud API
    """

    _client_id = None
    _client_secret = None
    _username = None
    _password = None
    _base_url = BASE_URL
    _access_token_url_ = ACCESS_URL
    _refresh_token_url = REFRESH_URL
    _proxies = None

    _token = None
    _token_expiration_time = None
    _refresh_token_expiration_time = None

    _verify_ssl = True
    _session = None

    def __init__(self, client_id, client_secret, username, password):
        """Instantiate a new iland.Api object.

        :param client_id: the client identifier
        :param client_secret: the client secret
        :param username: the iland cloud username
        :param password: the iland cloud password
        :return: Api Object
        """
        self._client_id = client_id
        self._client_secret = client_secret
        self._username = username
        self._password = password
        self._session = requests.Session()

    @property
    def _access_token_url(self):
        return self._access_token_url_

    @_access_token_url.setter
    def _access_token_url(self, access_token_url):
        if access_token_url is not None:
            self._access_token_url_ = access_token_url
            self._refresh_token_url = access_token_url + '?refresh=1'

    def _get_access_token(self):

        if self._valid_token():
            return self._token

        LOG.info("SSO Request %s" % self._access_token_url)

        params = {'client_id': self._client_id,
                  'client_secret': self._client_secret,
                  'username': self._username,
                  'password': self._password,
                  'grant_type': 'password'}
        r = self._session.post(
            self._access_token_url, data=params, verify=self._verify_ssl)
        json_payload = json.loads(r.content.decode('ascii'))
        if r.status_code not in [200, 201, 202]:
            raise UnauthorizedException(json_payload)
        self._token = json_payload
        self._set_token_expiration_time()
        return self._token

    def _set_token_expiration_time(self):
        time_buffer = 10
        now = int(round(time.time() * 1000))
        self._token_expiration_time = \
            ((self._token['expires_in'] - time_buffer) * 1000) + now
        self._refresh_token_expiration_time = \
            ((self._token['refresh_expires_in'] - time_buffer) * 1000) + now

    def _refresh_token(self):
        if not self._valid_token():
            if self._valid_refresh_token():
                LOG.info("SSO Request %s" % self._refresh_token_url)
                params = {'client_id': self._client_id,
                          'client_secret': self._client_secret,
                          'grant_type': 'refresh_token',
                          'refresh_token': self._token['refresh_token']
                          }
                r = self._session.post(self._refresh_token_url, data=params,
                                       verify=self._verify_ssl)
                json_payload = json.loads(r.content.decode('ascii'))
                if r.status_code not in [200, 201, 202]:
                    self._token = None
                    self._get_access_token()
                self._token = json_payload
                self._set_token_expiration_time()
            else:
                self._get_access_token()
        return self._token

    def _valid_refresh_token(self):
        if self._token is not None:
            now = int(round(time.time() * 1000))
            return now < self._refresh_token_expiration_time
        return False

    def _valid_token(self):
        if self._token is not None:
            now = int(round(time.time() * 1000))
            return now < self._token_expiration_time
        return False

    def _get_access_token_string(self):
        token_string = self._get_access_token()['access_token']
        return token_string

    def _do_request(self, rpath, verb='GET', form_data=None, headers=None,
                    timeout=None):
        self._refresh_token()
        data = None
        if form_data is not None:
            data = json.dumps(form_data, ensure_ascii=False).encode("UTF8")
        url = self._base_url + rpath

        LOG.info("Request %s rpath %s" % (verb, url))

        default_headers = {
            'Authorization': 'Bearer %s' % self._get_access_token_string(),
            'Accept': 'application/vnd.ilandcloud.api.v1.0+json'
        }
        if verb in ('PUT', 'POST'):
            default_headers[
                'Content-Type'] = 'application/json'

        merged_headers = default_headers.copy()

        if headers and isinstance(headers, dict):
            for header, value in headers.items():
                # don't allow overriding of our default headers
                if header in default_headers:
                    LOG.warning("Header '%s' can't be overridden" % header)
                else:
                    merged_headers[header] = value

        request_params = {
            'headers': merged_headers,
            'verify':  self._verify_ssl,
            'timeout': timeout
        }

        if verb in ('PUT', 'POST'):
            request_params['data'] = data

        if self._proxies and isinstance(self._proxies, dict):
            request_params['proxies'] = self._proxies

        if verb == 'GET':
            r = self._session.get(url, **request_params)
        elif verb == 'PUT':
            r = self._session.put(url, **request_params)
        elif verb == 'POST':
            r = self._session.post(url, **request_params)
        elif verb == 'DELETE':
            r = self._session.delete(url, **request_params)
        else:
            raise ApiException({'message': 'Unsupported HTTP verb %s' % verb})

        try:
            json_obj = r.json()
        except ValueError:
            raise ApiException(r.content)
        if r.status_code not in [200, 201, 202, 204]:
            raise ApiException(json_obj)
        return json_obj

    def get(self, rpath, headers=None, timeout=None):
        """ Perform a GET request against the iland cloud API given its
        resource path.

        `iland.Api` will refresh the access token if non valid.

        :param rpath: the resource path as a Python builtin String object
        :param headers: an optional dictionary of http headers to send with \
                        the request
        :raises: ApiException: API requests returns an error
        :raises: UnauthorizedException: credentials / grants invalids
        :return: a JSON Object or a list of JSON Objects.
        """
        return self._do_request(rpath, headers=headers, timeout=timeout)

    def put(self, rpath, form_data=None, headers=None, timeout=None):
        """ Perform a PUT request against the iland cloud API given its
        resource path.

        `iland.Api` will refresh the access token if non valid.

        :param rpath: the resource path as a Python builtin String object
        :param form_data: a Python builtin dict object
        :param headers: an optional dictionary of http headers to send with \
                        the request
        :raises: ApiException: API requests returns an error
        :raises: UnauthorizedException: credentials / grants invalids
        :return: a JSON Object or a list of JSON Objects.
        """
        return self._do_request(rpath, verb='PUT', form_data=form_data,
                                headers=headers)

    def post(self, rpath, form_data=None, headers=None, timeout=None):
        """ Perform a POST request against the iland cloud API given its
        resource path.

        `iland.Api` will refresh the access token if non valid.

        :param rpath: the resource path as a Python builtin String object
        :param form_data: a Python builtin dict object
        :param headers: an optional dictionary of http headers to send with \
                        the request
        :raises: ApiException: API requests returns an error
        :raises: UnauthorizedException: credentials / grants invalids
        :return: a JSON Object or a list of JSON Objects.
        """
        return self._do_request(rpath, verb='POST', form_data=form_data,
                                headers=headers)

    def delete(self, rpath, headers=None, timeout=None):
        """ Perform a DELETE request against the iland cloud API given its
        resource path.

        `iland.Api` will refresh the access token if non valid.

        :param rpath:  the resource path as a Python builtin String object
        :param headers: an optional dictionary of http headers to send with \
                        the request
        :raises: ApiException: API requests returns an error
        :raises: UnauthorizedException: credentials / grants invalids
        :return: a JSON Object or a list of JSON Objects.
        """
        return self._do_request(rpath, verb='DELETE', headers=headers)

    def get_access_token(self):
        """ Returns the access token in use for this session.

        This method is exposed in case you are interested in managing the
        token life cycle yourself. `iland.Api` will refresh the token on your
        behalf while performing queries.

        :raises: UnauthorizedException: credentials / grants invalids
        :return: JSON Object containing the actual access token
        """
        return self._get_access_token()

    def refresh_access_token(self):
        """ Refresh token if token is not valid: None or expired.

        This method is exposed in case you are interested in managing the
        token life cycle yourself. `iland.Api` will refresh the token on your
        behalf while performing queries.

        :raises: UnauthorizedException: credentials / grants invalids
        :return: JSON Object containing the actual access token
        """
        return self._refresh_token()

    def login(self):
        """ Requests an access token.

        This method is exposed in case you are interested in managing the
        token life cycle yourself. `iland.Api` will refresh the token on your
        behalf while performing queries.

        :raises: UnauthorizedException: credentials / grants invalids
        :return: JSON Object containing the actual access token
        """
        return self._get_access_token()
